# -*- coding: utf-8 -*-
import re
import glob
import os

from pyparsing import (
    Regex, Literal, White, Word, alphanums, CharsNotIn, Forward, Group,
    Optional, OneOrMore, ZeroOrMore, pythonStyleComment, lineno, LineStart, LineEnd
)

from amplify.agent.context import context

__author__ = "Mike Belov"
__copyright__ = "Copyright (C) 2015, Nginx Inc. All rights reserved."
__credits__ = ["Mike Belov", "Andrei Belov", "Ivan Poluyanov", "Oleg Mamontov", "Andrew Alexeev"]
__license__ = ""
__maintainer__ = "Mike Belov"
__email__ = "dedm@nginx.com"

tokens_cache = {}

def set_line_number(string, location, tokens):
    if len(tokens) == 1:
        line_number = lineno(location, string)
        tokens_cache[tokens[0]] = line_number
        tokens.line_number = line_number
    else:
        for item in tokens:
            tokens.line_number = tokens_cache.get(item)

class NginxConfigParser(object):
    """
    Nginx config parser based on https://github.com/fatiherikli/nginxparser
    Parses single file into json structure
    """

    max_size = 10*1024*1024  # 10 mb

    # line starts/ends
    line_start = LineStart().suppress()
    line_end = LineEnd().suppress()

    # constants
    left_brace = Literal("{").suppress()
    left_parentheses = Literal("(").suppress()
    right_brace = Literal("}").suppress()
    right_parentheses = Literal(")").suppress()
    semicolon = Regex(r";").suppress()
    space = White().suppress()

    # keys
    if_key = Literal("if").setParseAction(set_line_number)
    if_value = Regex(r'\(.*\)').setParseAction(set_line_number)

    key = ~if_key + Word(alphanums + "_/-+.*").setParseAction(set_line_number)
    value = CharsNotIn("{};").setParseAction(set_line_number)

    # modifier for location uri [ = | ~ | ~* | ^~ ]
    modifier = Literal("=") | Literal("~*") | Literal("~") | Literal("^~")

    # rules
    assignment = (
        key + Optional(space + value)
        + semicolon
    ).setParseAction(set_line_number)

    block = Forward()
    block << Group(
        (
            Group(if_key + Optional(space) + if_value + Optional(space)) |
            Group(
                key + Optional(space + modifier) + Optional(space) +
                Optional(value) + Optional(space) + Optional(value)
            )
        ).setParseAction(set_line_number)
        + left_brace
        + Group(ZeroOrMore(Group(assignment) | block).setParseAction(set_line_number)).setParseAction(set_line_number)
        + right_brace
    )

    script = OneOrMore(Group(assignment) | block).ignore(pythonStyleComment)

    def __init__(self, filename='/etc/nginx/nginx.conf'):
        global tokens_cache

        self.filename = filename
        self.files = {}  # to prevent cycle files and line indexing
        self.index = []  # stores index for all sections (points to file number and line number)
        self.folder = '/'.join(self.filename.split('/')[:-1])  # stores path to folder with main config
        self.errors = []

        tokens_cache = {}
        
        # parse
        self.tree = self.__logic_parse(self.__pyparse(self.filename))

    def __pyparse(self, path):
        """
        Loads and parses all files

        :param path: file path (can contain *)
        """

        # resolve local paths
        if not path.startswith('/'):
            path = '%s/%s' % (self.folder, path)
            path = path.replace('"', '')

        # load all files
        current_files = []
        if '*' in path:
            for filename in glob.glob(path):
                current_files.append(filename)
        else:
            current_files.append(path)

        # open each include and parse it
        result = {}

        for filename in current_files:
            if filename not in self.files:
                file_index = len(self.files)
                self.files[filename] = {
                    'index': file_index,
                    'lines': 0,
                    'size': 0,
                    'mtime': 0
                }

                try:
                    size = os.path.getsize(filename)
                    mtime = int(os.path.getmtime(filename))

                    self.files[filename]['size'] = size
                    self.files[filename]['mtime'] = mtime

                    if size > self.max_size:
                        self.errors.append('failed to load %s due to: too large %s bytes' % (filename, size))
                        continue
                    source = open(filename).read()
                    lines_count = source.count('\n')

                    self.files[filename]['lines'] = lines_count
                except Exception, e:
                    exception_name = e.__class__.__name__
                    message = 'failed to load %s due to: %s' % (filename, exception_name)
                    self.errors.append(message)
                    context.log.error(message)
                    context.log.debug('additional info:', exc_info=True)
                    continue

                # check that file contains some information (not commented)
                all_lines_commented = True
                for line in source.split('\n'):
                    if line and not(line.startswith('#')):
                        all_lines_commented = False
                        break
                if all_lines_commented:
                    continue

                try:
                    parsed = self.script.parseString(source, parseAll=True)
                except Exception, e:
                    exception_name = e.__class__.__name__
                    message = 'failed to parse %s due to %s' % (filename, exception_name)
                    self.errors.append(message)
                    context.log.error(message)
                    context.log.debug('additional info:', exc_info=True)
                    continue

                result[file_index] = list(parsed)

        return result

    def __logic_parse(self, files, result=None):
        """
        Parses input files and updates result dict

        :param files: dict of files from pyparsing
        :return: dict of config tree
        """
        if result is None:
            result = {}

        for file_index, rows in files.iteritems():
            while len(rows):
                row = rows.pop(0)
                row_as_list = row.asList()
                
                if isinstance(row_as_list[0], list):
                    # this is a new key
                    key_bucket, value_bucket = row
                    key = key_bucket[0]

                    if len(key_bucket) == 1:
                        # simple key, with one param
                        subtree_indexed = self.__idx_save(
                            self.__logic_parse({file_index: row[1]}),
                            file_index, row.line_number
                        )
                        if key == 'server':
                            # work with servers
                            if key in result:
                                result[key].append(subtree_indexed)
                            else:
                                result[key] = [subtree_indexed]
                        else:
                            result[key] = subtree_indexed
                    else:
                        # compound key (for locations and upstreams for example)

                        # remove all redundant spaces
                        parts = filter(lambda x: x, ' '.join(key_bucket[1:]).split(' '))
                        sub_key = ' '.join(parts)

                        subtree_indexed = self.__idx_save(
                            self.__logic_parse({file_index: row[1]}),
                            file_index, row.line_number
                        )

                        if key in result:
                            result[key][sub_key] = subtree_indexed
                        else:
                            result[key] = {sub_key: subtree_indexed}
                else:
                    # can be just an assigment, without value
                    if len(row) == 2:
                        key, value = row
                    else:
                        key, value = row[0], ''

                    # transform multiline values to single one
                    if """\'""" in value or """\n""" in value:
                        value = re.sub(r"\'\n\s*\'", '', value)
                        value = re.sub(r"\'", "'", value)

                    if key == 'log_format':
                        # work with log formats
                        gwe = re.match("([\w\d_-]+)\s+'(.+)'", value)
                        if gwe:
                            format_name, format_value = gwe.group(1), gwe.group(2)

                            indexed_value = self.__idx_save(format_value, file_index, row.line_number)
                            if key in result:
                                result[key][format_name] = indexed_value
                            else:
                                result[key] = {format_name: indexed_value}
                    elif key == 'include':
                        included_files = self.__pyparse(value)
                        self.__logic_parse(included_files, result=result)
                    else:
                        indexed_value = self.__idx_save(value, file_index, row.line_number)

                        # simple key-value
                        if key in result:
                            stored_value = result[key]
                            if isinstance(stored_value, list):
                                result[key].append(indexed_value)
                            else:
                                result[key] = [stored_value, indexed_value]
                        else:
                            result[key] = indexed_value
        return result

    def __idx_save(self, value, file_index, line):
        new_index = len(self.index)
        self.index.append((file_index, line))
        return value, new_index
    
    def simplify(self, tree=None):
        """
        returns tree without index references
        can be used for debug/pretty output

        :return: dict of self.tree without index positions
        """
        result = {}

        if tree is None:
            tree = self.tree

        if isinstance(tree, dict):
            for key, value in tree.iteritems():
                if isinstance(value, dict):
                    result[key] = self.simplify(tree=value)
                elif isinstance(value, tuple):
                    subtree, reference = value
                    if isinstance(subtree, dict):
                        result[key] = self.simplify(tree=subtree)
                    elif isinstance(subtree, list):
                        result[key] = map(lambda x: self.simplify(tree=x), subtree)
                    else:
                        result[key] = subtree
                elif isinstance(value, list):
                    result[key] = map(lambda x: self.simplify(tree=x), value)
        elif isinstance(tree, tuple):
            subtree, reference = tree
            if isinstance(subtree, dict):
                return self.simplify(tree=subtree)
            elif isinstance(subtree, list):
                return map(lambda x: self.simplify(tree=x), subtree)
            else:
                return subtree
        elif isinstance(tree, list):
            return map(lambda x: self.simplify(tree=x), tree)

        return result

